@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
  * USA
  */
 
@@ -90,14 +90,14 @@ gsf_input_textline_finalize (GObject *obj)
 }
 
 static GsfInput *
-gsf_input_textline_dup (GsfInput *src_input, GError **err)
+gsf_input_textline_dup (GsfInput *src_input, G_GNUC_UNUSED GError **err)
 {
 	GsfInputTextline const *src = (GsfInputTextline *)src_input;
 	GsfInputTextline *dst = g_object_new (GSF_INPUT_TEXTLINE_TYPE, NULL);
 
-	(void)err;
 	dst->source = src->source;
 	g_object_ref (G_OBJECT (dst->source));
+	gsf_input_set_size (GSF_INPUT (dst), gsf_input_size (src_input));
 
 	return GSF_INPUT (dst);
 }
@@ -152,7 +152,7 @@ GSF_CLASS (GsfInputTextline, gsf_input_textline,
  * @input :
  *
  * A utility routine to read things line by line from the underlying source.
- * Trailing newlines and carriage returns are stipped, and the resultant buffer
+ * Trailing newlines and carriage returns are stripped, and the resultant buffer
  * can be edited.
  *
  * returns the string read, or NULL on eof.
@@ -239,7 +239,7 @@ gsf_input_textline_ascii_gets (GsfInputTextline *textline)
  * @input :
  *
  * A utility routine to read things line by line from the underlying source.
- * Trailing newlines and carriage returns are stipped, and the resultant buffer
+ * Trailing newlines and carriage returns are stripped, and the resultant buffer
  * can be edited.
  *
  * returns the string read, or NULL on eof.
