@@ -2,7 +2,7 @@
 /*
  * gsf-libxml.c :
  *
- * Copyright (C) 2002-2004 Jody Goldberg (jody@gnome.org)
+ * Copyright (C) 2002-2005 Jody Goldberg (jody@gnome.org)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of version 2.1 of the GNU Lesser General Public
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
  * USA
  */
 
@@ -30,9 +30,165 @@
 #include <math.h>
 #include <string.h>
 
+static GObjectClass *parent_class;
+
+static gint
+glade_enum_from_string (GType type, const char *string)
+{
+    GEnumClass *eclass;
+    GEnumValue *ev;
+    gchar *endptr;
+    gint ret = 0;
+
+    ret = strtoul(string, &endptr, 0);
+    if (endptr != string) /* parsed a number */
+	return ret;
+
+    eclass = g_type_class_ref(type);
+    ev = g_enum_get_value_by_name(eclass, string);
+    if (!ev) ev = g_enum_get_value_by_nick(eclass, string);
+    if (ev)  ret = ev->value;
+
+    g_type_class_unref(eclass);
+
+    return ret;
+}
+
+static guint
+glade_flags_from_string (GType type, const char *string)
+{
+    GFlagsClass *fclass;
+    gchar *endptr, *prevptr;
+    guint i, j, ret = 0;
+    char *flagstr;
+
+    ret = strtoul(string, &endptr, 0);
+    if (endptr != string) /* parsed a number */
+	return ret;
+
+    fclass = g_type_class_ref(type);
+
+
+    flagstr = g_strdup (string);
+    for (ret = i = j = 0; ; i++) {
+	gboolean eos;
+
+	eos = flagstr [i] == '\0';
+	
+	if (eos || flagstr [i] == '|') {
+	    GFlagsValue *fv;
+	    const char  *flag;
+	    gunichar ch;
+
+	    flag = &flagstr [j];
+            endptr = &flagstr [i];
+
+	    if (!eos) {
+		flagstr [i++] = '\0';
+		j = i;
+	    }
+
+            /* trim spaces */
+	    for (;;)
+	      {
+		ch = g_utf8_get_char (flag);
+		if (!g_unichar_isspace (ch))
+		  break;
+		flag = g_utf8_next_char (flag);
+	      }
+
+	    while (endptr > flag)
+	      {
+		prevptr = g_utf8_prev_char (endptr);
+		ch = g_utf8_get_char (prevptr);
+		if (!g_unichar_isspace (ch))
+		  break;
+		endptr = prevptr;
+	      }
+
+	    if (endptr > flag)
+	      {
+		*endptr = '\0';
+		fv = g_flags_get_value_by_name (fclass, flag);
+
+		if (!fv)
+		  fv = g_flags_get_value_by_nick (fclass, flag);
+
+		if (fv)
+		  ret |= fv->value;
+		else
+		  g_warning ("Unknown flag: '%s'", flag);
+	      }
+
+	    if (eos)
+		break;
+	}
+    }
+    
+    g_free (flagstr);
+
+    g_type_class_unref(fclass);
+
+    return ret;
+}
+
+gboolean
+gsf_xml_gvalue_from_str (GValue *res, GType t, char const *str)
+{
+	g_return_val_if_fail (res != NULL, FALSE);
+	g_return_val_if_fail (str != NULL, FALSE);
+
+	g_value_init (res, t);
+	switch (t) {
+	case G_TYPE_CHAR:
+		g_value_set_char (res, str[0]);
+		break;
+	case G_TYPE_UCHAR:
+		g_value_set_uchar (res, (guchar)str[0]);
+		break;
+	case G_TYPE_BOOLEAN:
+		g_value_set_boolean (res, 
+			g_ascii_tolower (str[0]) == 't' ||
+			g_ascii_tolower (str[0]) == 'y' ||
+			strtol (str, NULL, 0));
+		break;
+	case G_TYPE_INT:
+		g_value_set_int (res, strtol (str, NULL, 0));
+		break;
+	case G_TYPE_UINT:
+		g_value_set_uint (res, strtoul (str, NULL, 0));
+		break;
+	case G_TYPE_LONG:
+		g_value_set_long (res, strtol (str, NULL, 0));
+		break;
+	case G_TYPE_ULONG:
+		g_value_set_ulong (res, strtoul (str, NULL, 0));
+		break; 
+	case G_TYPE_ENUM:
+		g_value_set_enum (res, glade_enum_from_string (t, str));
+		break;
+	case G_TYPE_FLAGS:
+		g_value_set_flags (res, glade_flags_from_string (t, str));
+		break;
+	case G_TYPE_FLOAT:
+		g_value_set_float (res, g_strtod (str, NULL));
+		break;
+	case G_TYPE_DOUBLE:
+		g_value_set_double (res, g_strtod (str, NULL));
+		break;
+	case G_TYPE_STRING:
+		g_value_set_string (res, str);
+		break;
+
+	default:
+		return FALSE;
+	}
+	return TRUE;
+}
+
 /* Note: libxml erroneously declares the length argument as int.  */
 static int
-gsf_libxml_read (void *context, char *buffer, int len)
+gsf_libxml_read (void *context, guint8 *buffer, int len)
 {
 	gsf_off_t remaining = gsf_input_remaining ((GsfInput *)context);
 	guint8* res;
@@ -64,13 +220,13 @@ gsf_libxml_close (void *context)
 static xmlParserCtxtPtr
 gsf_xml_parser_context_full (GsfInput *input, xmlSAXHandlerPtr sax, gpointer user)
 {
-	GsfInputGZip *gzip;
+	GsfInput *gzip;
 
 	g_return_val_if_fail (GSF_IS_INPUT (input), NULL);
 
 	gzip = gsf_input_gzip_new (input, NULL);
 	if (gzip != NULL)
-		input = GSF_INPUT (gzip);
+		input = gzip;
 	else
 		g_object_ref (G_OBJECT (input));
 
@@ -124,7 +280,7 @@ gsf_xml_output_buffer_new (GsfOutput *output,
 }
 
 int
-gsf_xmlDocFormatDump (GsfOutput *output, xmlDocPtr cur, const char * encoding,
+gsf_xmlDocFormatDump (GsfOutput *output, xmlDocPtr cur, char const *encoding,
 		      gboolean format)
 {
 	xmlOutputBufferPtr buf;
@@ -166,6 +322,7 @@ struct _GsfXMLInDoc {
 	GsfXMLInNode	*root;
 	GsfXMLInNS	*ns;
 	GPtrArray	*ns_by_id;
+	GsfXMLInUnknownFunc	unknown_handler;
 };
 
 typedef struct {
@@ -179,17 +336,72 @@ typedef struct {
 	GSList *elem;
 } GsfXMLInNodeGroup;
 
+static char const *
+node_name (GsfXMLInNode const *node)
+{
+	return (node->name != NULL) ? node->name : "{catch all)}";
+}
+
+static gboolean
+lookup_child (GsfXMLIn *state, GsfXMLInNS const *default_ns,
+	      GsfXMLInNSInstance *inst,
+	      GSList *groups, xmlChar const *name, xmlChar const **attrs)
+{
+	GsfXMLInNodeGroup  *group;
+	GsfXMLInNode	   *node;
+	GSList *elem, *ptr;
+	char const *tmp;
+
+	for (ptr = groups ; ptr != NULL ; ptr = ptr->next) {
+		group = ptr->data;
+		/* does the namespace match */
+		if (group->ns != NULL && group->ns != default_ns) {
+			g_return_val_if_fail (state->ns_by_id->len > group->ns->ns_id, FALSE);
+			inst = g_ptr_array_index (state->ns_by_id, group->ns->ns_id);
+			if (inst == NULL || 0 != strncmp (name, inst->tag, inst->taglen))
+				continue;
+			tmp = name + inst->taglen;
+		} else {
+#if 0
+			g_return_if_fail (state->ns_by_id->len > group->ns->ns_id);
+			inst = g_ptr_array_index (state->ns_by_id, group->ns->ns_id);
+			g_warning ("accepted ns = '%s' looking for '%s'", inst->tag, name);
+#endif
+			tmp = name;
+		}
+		for (elem = group->elem ; elem != NULL ; elem = elem->next) {
+			node = elem->data;
+			if (node->name == NULL || !strcmp (tmp, node->name)) {
+				if (node->has_content == GSF_XML_CONTENT &&
+				    state->content->len > 0) {
+					g_warning ("too lazy to support nested unshared content for now.  We'll add it for 2.0");
+				}
+				state->state_stack = g_slist_prepend (state->state_stack,
+								      (gpointer)state->node);
+				state->ns_stack = g_slist_prepend (state->ns_stack,
+								   (gpointer)state->default_ns);
+				state->node = node;
+				state->default_ns = default_ns;
+				if (node->start != NULL)
+					node->start (state, attrs);
+				return TRUE;
+			}
+		}
+	}
+	return FALSE;
+}
+
 static void
 gsf_xml_in_start_element (GsfXMLIn *state, xmlChar const *name, xmlChar const **attrs)
 {
-	GSList *ptr, *elem;
 	GsfXMLInNSInstance *inst;
-	GsfXMLInNodeGroup  *group;
-	GsfXMLInNode	   *node;
 	GsfXMLInNS const   *ns, *default_ns = state->default_ns;
+	GsfXMLInNode const *node;
 	xmlChar const **ns_ptr;
+	GSList *ptr;
 	char const *tmp;
 	int i;
+	gboolean check_unknown_handler = TRUE;
 
 	/* Scan for namespace declarations.  Yes it is ugly to have the api
 	 * flag that its children can declare namespaces. However, given that a
@@ -220,9 +432,15 @@ gsf_xml_in_start_element (GsfXMLIn *state, xmlChar const *name, xmlChar const **
 					inst->tag    = g_strconcat (ns_ptr[0] + 6, ":", NULL);
 					inst->taglen = strlen (inst->tag);
 					inst->ref_count = 1;
+					g_hash_table_insert (state->ns_prefixes, inst->tag, inst);
+
 					if (ns[i].ns_id >= state->ns_by_id->len)
 						g_ptr_array_set_size  (state->ns_by_id, ns[i].ns_id+1);
-					g_ptr_array_index (state->ns_by_id, ns[i].ns_id) = inst;
+					if (g_ptr_array_index (state->ns_by_id, ns[i].ns_id)) {
+						g_warning ("Damn.  Someone just declared the same namespace '%s' with a different prefix '%s'",
+							   ns[i].uri, inst->tag);
+					} else
+						g_ptr_array_index (state->ns_by_id, ns[i].ns_id) = inst;
 				} else
 					inst->ref_count++;
 				break;
@@ -230,63 +448,40 @@ gsf_xml_in_start_element (GsfXMLIn *state, xmlChar const *name, xmlChar const **
 		}
 	}
 
-	for (ptr = state->node->groups ; ptr != NULL ; ptr = ptr->next) {
-		group = ptr->data;
-		/* does the namespace match */
-		if (group->ns != NULL && group->ns != default_ns) {
-			g_return_if_fail (state->ns_by_id->len > group->ns->ns_id);
-			inst = g_ptr_array_index (state->ns_by_id, group->ns->ns_id);
-			if (0 != strncmp (name, inst->tag, inst->taglen))
-				continue;
-			tmp = name + inst->taglen;
-		} else {
-#if 0
-			g_return_if_fail (state->ns_by_id->len > group->ns->ns_id);
-			inst = g_ptr_array_index (state->ns_by_id, group->ns->ns_id);
-			g_warning ("accepted ns = '%s' looking for '%s'", inst->tag, name);
-#endif
-			tmp = name;
-		}
-		for (elem = group->elem ; elem != NULL ; elem = elem->next) {
-			node = elem->data;
-			if (node->name != NULL && !strcmp (tmp, node->name)) {
-				if (node->has_content == GSF_XML_CONTENT &&
-				    state->content->len > 0) {
-					g_warning ("too lazy to support nested unshared content for now.  We'll add it for 2.0");
-				}
-				state->state_stack = g_slist_prepend (state->state_stack,
-								      (gpointer)state->node);
-				state->ns_stack = g_slist_prepend (state->ns_stack,
-								   (gpointer)state->default_ns);
-				state->node = node;
-				state->default_ns = default_ns;
-				if (node->start != NULL)
-					node->start (state, attrs);
-				return;
-			}
-		}
+retry_after_unknown :
+	if (lookup_child (state, default_ns, inst, state->node->groups, name, attrs))
+		return;
+
+	/* useful for <Data><b><i><u></u></i></b></Data> where all of the markup can nest */
+	node = state->node;
+	ptr = state->state_stack;
+	for (; ptr != NULL && node->share_children_with_parent; ptr = ptr->next) {
+		node = ptr->data;
+		if (lookup_child (state, default_ns, inst, node->groups, name, attrs))
+			return;
 	}
 
+	if (check_unknown_handler) {
+		check_unknown_handler = FALSE; /* only loop once */
+		if (state->doc->unknown_handler != NULL &&
+		    (state->doc->unknown_handler) (state, name, attrs))
+			goto retry_after_unknown;
+	}
 	if (state->unknown_depth++)
 		return;
-	g_warning ("Unexpected element '%s' in state %s.", name, state->node->name);
-	{
-		GSList *ptr;
-		GsfXMLInNode *node;
-		ptr = state->state_stack = g_slist_reverse (state->state_stack);
-		for (;ptr != NULL && ptr->next != NULL; ptr = ptr->next) {
-			node = ptr->data;
-			if (node != NULL) {
-#ifdef __GNUC__
-#warning if we really want this do we also want namespaces ?
-#endif
-				g_print ("%s", node->name);
-				if (ptr->next != NULL && ptr->next->data != NULL)
-					g_print (" -> ");
-			}
+	g_warning ("Unexpected element '%s' in state %s.", name, node_name (state->node));
+
+	ptr = state->state_stack = g_slist_reverse (state->state_stack);
+	for (;ptr != NULL && ptr->next != NULL; ptr = ptr->next) {
+		node = ptr->data;
+		if (node != NULL) {
+/* FIXME FIXME FIXME if we really want this do we also want namespaces ? */
+			g_print ("%s", node_name (node));
+			if (ptr->next != NULL && ptr->next->data != NULL)
+				g_print (" -> ");
 		}
-		state->state_stack = g_slist_reverse (state->state_stack);
 	}
+	state->state_stack = g_slist_reverse (state->state_stack);
 }
 
 static void
@@ -321,9 +516,8 @@ gsf_xml_in_characters (GsfXMLIn *state, const xmlChar *chars, int len)
 }
 
 static xmlEntityPtr
-gsf_xml_in_get_entity (GsfXMLIn *state, const xmlChar *name)
+gsf_xml_in_get_entity (G_GNUC_UNUSED GsfXMLIn *state, const xmlChar *name)
 {
-	(void)state;
 	return xmlGetPredefinedEntity (name);
 }
 
@@ -337,7 +531,7 @@ gsf_xml_in_start_document (GsfXMLIn *state)
 	state->default_ns  = NULL;
 	state->ns_by_id    = g_ptr_array_new ();
 	state->ns_prefixes = g_hash_table_new_full (g_str_hash, g_str_equal,
-		g_free, g_free);
+		NULL, g_free);
 }
 
 static void
@@ -357,66 +551,63 @@ gsf_xml_in_end_document (GsfXMLIn *state)
 }
 
 static void
-gsf_xml_in_warning (GsfXMLIn *state, const char *msg, ...)
+gsf_xml_in_warning (G_GNUC_UNUSED GsfXMLIn *state, char const *msg, ...)
 {
 	va_list args;
 
-	(void)state;
 	va_start (args, msg);
 	g_logv ("XML", G_LOG_LEVEL_WARNING, msg, args);
 	va_end (args);
 }
 
 static void
-gsf_xml_in_error (GsfXMLIn *state, const char *msg, ...)
+gsf_xml_in_error (G_GNUC_UNUSED GsfXMLIn *state, char const *msg, ...)
 {
 	va_list args;
 
-	(void)state;
 	va_start (args, msg);
 	g_logv ("XML", G_LOG_LEVEL_CRITICAL, msg, args);
 	va_end (args);
 }
 
 static void
-gsf_xml_in_fatal_error (GsfXMLIn *state, const char *msg, ...)
+gsf_xml_in_fatal_error (G_GNUC_UNUSED GsfXMLIn *state, char const *msg, ...)
 {
 	va_list args;
 
-	(void)state;
 	va_start (args, msg);
 	g_logv ("XML", G_LOG_LEVEL_ERROR, msg, args);
 	va_end (args);
 }
 
 static xmlSAXHandler gsfXMLInParser = {
-	0, /* internalSubset */
-	0, /* isStandalone */
-	0, /* hasInternalSubset */
-	0, /* hasExternalSubset */
-	0, /* resolveEntity */
+	NULL, /* internalSubset */
+	NULL, /* isStandalone */
+	NULL, /* hasInternalSubset */
+	NULL, /* hasExternalSubset */
+	NULL, /* resolveEntity */
 	(getEntitySAXFunc)gsf_xml_in_get_entity, /* getEntity */
-	0, /* entityDecl */
-	0, /* notationDecl */
-	0, /* attributeDecl */
-	0, /* elementDecl */
-	0, /* unparsedEntityDecl */
-	0, /* setDocumentLocator */
+	NULL, /* entityDecl */
+	NULL, /* notationDecl */
+	NULL, /* attributeDecl */
+	NULL, /* elementDecl */
+	NULL, /* unparsedEntityDecl */
+	NULL, /* setDocumentLocator */
 	(startDocumentSAXFunc)gsf_xml_in_start_document, /* startDocument */
 	(endDocumentSAXFunc)gsf_xml_in_end_document, /* endDocument */
 	(startElementSAXFunc)gsf_xml_in_start_element, /* startElement */
 	(endElementSAXFunc)gsf_xml_in_end_element, /* endElement */
-	0, /* reference */
+	NULL, /* reference */
 	(charactersSAXFunc)gsf_xml_in_characters, /* characters */
-	0, /* ignorableWhitespace */
-	0, /* processingInstruction */
-	0, /* comment */
+	NULL, /* ignorableWhitespace */
+	NULL, /* processingInstruction */
+	NULL, /* comment */
 	(warningSAXFunc)gsf_xml_in_warning, /* warning */
 	(errorSAXFunc)gsf_xml_in_error, /* error */
 	(fatalErrorSAXFunc)gsf_xml_in_fatal_error, /* fatalError */
-	0, /* getParameterEntity */
-	0, /* cdataBlock */
-	0, /* externalSubset */
+	NULL, /* getParameterEntity */
+	NULL, /* cdataBlock */
+	NULL, /* externalSubset */
 	0
 #if LIBXML_VERSION >= 20600
 	,
@@ -444,8 +635,6 @@ GsfXMLInDoc *
 gsf_xml_in_doc_new (GsfXMLInNode *root, GsfXMLInNS *ns)
 {
 	GsfXMLInDoc  *doc;
-	GHashTable   *symbols;
-	GsfXMLInNode *tmp, *real_node, *node;
 	unsigned i;
 
 	if (root->parent_initialized)
@@ -464,10 +653,24 @@ gsf_xml_in_doc_new (GsfXMLInNode *root, GsfXMLInNS *ns)
 			g_ptr_array_index (doc->ns_by_id, ns[i].ns_id) = ns+i;
 		}
 	}
+	gsf_xml_in_doc_extend (doc, root);
+	return doc;
+}
 
-	symbols = g_hash_table_new (g_str_hash, g_str_equal);
-	for (node = root; node->id != NULL ; node++ ) {
-		g_return_val_if_fail (!node->parent_initialized, NULL);
+void
+gsf_xml_in_doc_extend (GsfXMLInDoc  *doc,
+		       GsfXMLInNode *extension_nodes)
+{
+	GsfXMLInNode *tmp, *real_node, *node;
+	GHashTable   *symbols = g_hash_table_new (g_str_hash, g_str_equal);
+
+	g_return_if_fail (doc != NULL);
+	g_return_if_fail (extension_nodes != NULL);
+
+	if (extension_nodes->parent_initialized)
+		return;
+	for (node = extension_nodes; node->id != NULL ; node++ ) {
+		g_return_if_fail (!node->parent_initialized);
 
 		tmp = g_hash_table_lookup (symbols, node->id);
 		if (tmp != NULL) {
@@ -476,7 +679,7 @@ gsf_xml_in_doc_new (GsfXMLInNode *root, GsfXMLInNS *ns)
 			    node->has_content != GSF_XML_NO_CONTENT || node->user_data.v_int != 0) {
 				g_warning ("ID '%s' has already been registered.\n"
 					   "The additional decls should not specify start,end,content,data", node->id);
-				return NULL;
+				return;
 			}
 			real_node = tmp;
 		} else {
@@ -508,7 +711,7 @@ gsf_xml_in_doc_new (GsfXMLInNode *root, GsfXMLInNS *ns)
 			group->elem = g_slist_prepend (group->elem, real_node);
 		} else if (strcmp (node->id, node->parent_id)) {
 			g_warning ("Parent ID '%s' unknown", node->parent_id);
-			return NULL;
+			return;
 		}
 
 		/* WARNING VILE HACK :
@@ -524,15 +727,28 @@ gsf_xml_in_doc_new (GsfXMLInNode *root, GsfXMLInNS *ns)
 	}
 
 	g_hash_table_destroy (symbols);
+}
 
-	return doc;
+/**
+ * gsf_xml_in_doc_set_unknown_handler:
+ * @doc : #GsfXMLInDoc
+ * @handler : The function to call
+ *
+ * Call the function @handler when an unexpected child node is found
+ **/
+void
+gsf_xml_in_doc_set_unknown_handler (GsfXMLInDoc *doc,
+				    GsfXMLInUnknownFunc handler)
+{
+	g_return_if_fail (doc != NULL);
+	doc->unknown_handler = handler;
 }
 
 /**
  * gsf_xml_in_doc_free :
  * @doc :
  *
- * Free up resources allocated by gsf_xml_in_doc_prep
+ * Free up resources
  **/
 void
 gsf_xml_in_doc_free (GsfXMLInDoc *doc)
@@ -560,29 +776,32 @@ gsf_xml_in_doc_free (GsfXMLInDoc *doc)
 }
 
 /**
- * gsf_xml_in_parse :
- * @state :
+ * gsf_xml_in_doc_parse :
+ * @doc :
  * @input :
+ * @user_state :
  *
  * Read an xml document from @input and parse based on the the descriptor in
- * @state::doc
+ * @doc
  *
  * returns FALSE on error
  **/
 gboolean
-gsf_xml_in_parse (GsfXMLIn *state, GsfInput *input)
+gsf_xml_in_doc_parse (GsfXMLInDoc *doc, GsfInput *input, gpointer user_state)
 {
-	xmlParserCtxt *ctxt;
+	xmlParserCtxt	*ctxt;
+	GsfXMLIn	 state;
 	gboolean res;
 
-	g_return_val_if_fail (state != NULL, FALSE);
-	g_return_val_if_fail (GSF_IS_INPUT (input), FALSE);
+	g_return_val_if_fail (doc != NULL, FALSE);
 
-	ctxt = gsf_xml_parser_context_full (input, &gsfXMLInParser, state);
+	ctxt = gsf_xml_parser_context_full (input, &gsfXMLInParser, &state);
+	if (ctxt == NULL)
+		return FALSE;
 
-	g_return_val_if_fail (ctxt != NULL, FALSE);
-
-	state->content = g_string_sized_new (128);
+	state.doc = doc;
+	state.user_state = user_state;
+	state.content = g_string_sized_new (128);
 	xmlParseDocument (ctxt);
 	res = ctxt->wellFormed;
 	xmlFreeParserCtxt (ctxt);
@@ -604,8 +823,11 @@ gsf_xml_in_parse (GsfXMLIn *state, GsfInput *input)
 char const *
 gsf_xml_in_check_ns (GsfXMLIn const *state, char const *str, unsigned int ns_id)
 {
-	GsfXMLInNSInstance *inst = g_ptr_array_index (state->ns_by_id, ns_id);
-
+	GsfXMLInNSInstance *inst;
+	if (state->ns_by_id->len <= ns_id)
+		return NULL;
+	if (NULL == (inst = g_ptr_array_index (state->ns_by_id, ns_id)))
+		return NULL;
 	if (strncmp (str, inst->tag, inst->taglen))
 		return NULL;
 	return str + inst->taglen;
@@ -624,8 +846,11 @@ gboolean
 gsf_xml_in_namecmp (GsfXMLIn const *state, char const *str,
 		    unsigned int ns_id, char const *name)
 {
-	GsfXMLInNSInstance *inst = g_ptr_array_index (state->ns_by_id, ns_id);
-
+	GsfXMLInNSInstance *inst;
+	if (state->ns_by_id->len <= ns_id)
+		return FALSE;
+	if (NULL == (inst = g_ptr_array_index (state->ns_by_id, ns_id)))
+		return FALSE;
 	if (strncmp (str, inst->tag, inst->taglen))
 		return FALSE;
 	return 0 == strcmp (name, str + inst->taglen);
@@ -643,6 +868,7 @@ struct _GsfXMLOut {
 	GObject	   base;
 
 	GsfOutput	 *output;
+	char		 *doc_type;
 	GSList		 *stack;
 	GsfXMLOutState state;
 	unsigned   	  indent;
@@ -656,13 +882,11 @@ typedef struct {
 static void
 gsf_xml_out_finalize (GObject *obj)
 {
-	GObjectClass *parent_class;
+	GsfXMLOut *xml = GSF_XML_OUT (obj);
 
-	/* already unwrapped */
+	g_free (xml->doc_type);
 
-	parent_class = g_type_class_peek (G_TYPE_OBJECT);
-	if (parent_class && parent_class->finalize)
-		parent_class->finalize (obj);
+	parent_class->finalize (obj);
 }
 
 static void
@@ -674,12 +898,14 @@ gsf_xml_out_init (GObject *obj)
 	xml->state  = GSF_XML_OUT_CHILD;
 	xml->indent = 0;
 	xml->needs_header = TRUE;
+	xml->doc_type = NULL;
 }
 
 static void
 gsf_xml_out_class_init (GObjectClass *gobject_class)
 {
 	gobject_class->finalize = gsf_xml_out_finalize;
+	parent_class = g_type_class_peek_parent (gobject_class);
 }
 
 GSF_CLASS (GsfXMLOut, gsf_xml_out,
@@ -696,6 +922,20 @@ gsf_xml_out_new (GsfOutput *output)
 	return xml;
 }
 
+/**
+ * gsf_xml_out_set_doc_type :
+ * @xml : #GsfXMLOut
+ * @type :
+ * 
+ * Store some optional some &lt;!DOCTYPE .. &gt; content
+ **/
+void
+gsf_xml_out_set_doc_type (GsfXMLOut *xml, char const *type)
+{
+	g_free (xml->doc_type);
+	xml->doc_type = g_strdup (type);
+}
+
 static inline void
 gsf_xml_out_indent (GsfXMLOut *xml)
 {
@@ -725,9 +965,11 @@ gsf_xml_out_start_element (GsfXMLOut *xml, char const *id)
 	g_return_if_fail (xml->state != GSF_XML_OUT_CONTENT);
 
 	if (xml->needs_header) {
-		static char const header[] =
+		static char const header0[] =
 			"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
-		gsf_output_write (xml->output, sizeof (header) - 1, header);
+		gsf_output_write (xml->output, sizeof (header0) - 1, header0);
+		if (xml->doc_type != NULL)
+			gsf_output_puts (xml->output, xml->doc_type);
 		xml->needs_header = FALSE;
 	}
 	if (xml->state == GSF_XML_OUT_NOCONTENT)
@@ -758,15 +1000,15 @@ gsf_xml_out_end_element (GsfXMLOut *xml)
 	xml->indent--;
 	switch (xml->state) {
 	case GSF_XML_OUT_NOCONTENT :
-	gsf_output_write (xml->output, 3, "/>\n");
-	break;
+		gsf_output_write (xml->output, 3, "/>\n");
+		break;
 
 	case GSF_XML_OUT_CHILD :
-	gsf_xml_out_indent (xml);
+		gsf_xml_out_indent (xml);
 	/* fall through */
 	case GSF_XML_OUT_CONTENT :
-	gsf_output_printf (xml->output, "</%s>\n", id);
-	};
+		gsf_output_printf (xml->output, "</%s>\n", id);
+	}
 	xml->state = GSF_XML_OUT_CHILD;
 	return id;
 }
@@ -784,7 +1026,8 @@ gsf_xml_out_simple_element (GsfXMLOut *xml, char const *id,
 			    char const *content)
 {
 	gsf_xml_out_start_element (xml, id);
-	gsf_xml_out_add_cstr (xml, NULL, content);
+	if (content != NULL)
+		gsf_xml_out_add_cstr (xml, NULL, content);
 	gsf_xml_out_end_element (xml);
 }
 
@@ -822,6 +1065,15 @@ gsf_xml_out_simple_float_element (GsfXMLOut *xml, char const *id,
 	gsf_xml_out_end_element (xml);
 }
 
+static void
+close_tag_if_neccessary (GsfXMLOut* xml)
+{
+	if (xml->state != GSF_XML_OUT_CONTENT) {
+		xml->state = GSF_XML_OUT_CONTENT;
+		gsf_output_write (xml->output, 1, ">");
+	}
+}
+
 /**
  * gsf_xml_out_add_cstr_unchecked :
  * @xml :
@@ -830,18 +1082,20 @@ gsf_xml_out_simple_float_element (GsfXMLOut *xml, char const *id,
  *
  * dump @val_utf8 to an attribute named @id without checking to see if the
  * content needs escaping.  A useful performance enhancement when the
- * application knows that structure of the content well.
+ * application knows that structure of the content well.  If @val_utf8 is NULL
+ * do nothing (no warning, no output)
  **/
 void
 gsf_xml_out_add_cstr_unchecked (GsfXMLOut *xml, char const *id,
 				char const *val_utf8)
 {
 	g_return_if_fail (xml != NULL);
-	g_return_if_fail (xml->state == GSF_XML_OUT_NOCONTENT);
+
+	if (val_utf8 == NULL)
+		return;
 
 	if (id == NULL) {
-		xml->state = GSF_XML_OUT_CONTENT;
-		gsf_output_write (xml->output, 1, ">");
+		close_tag_if_neccessary (xml);
 		gsf_output_write (xml->output, strlen (val_utf8), val_utf8);
 	} else
 		gsf_output_printf (xml->output, " %s=\"%s\"", id, val_utf8);
@@ -854,7 +1108,8 @@ gsf_xml_out_add_cstr_unchecked (GsfXMLOut *xml, char const *id,
  * @val_utf8 : a utf8 encoded string
  *
  * dump @val_utf8 to an attribute named @id or as the nodes content escaping
- * characters as necessary.
+ * characters as necessary.  If @val_utf8 is NULL do nothing (no warning, no
+ * output)
  **/
 void
 gsf_xml_out_add_cstr (GsfXMLOut *xml, char const *id,
@@ -863,9 +1118,13 @@ gsf_xml_out_add_cstr (GsfXMLOut *xml, char const *id,
 	guint8 const *cur   = val_utf8;
 	guint8 const *start = val_utf8;
 
+	g_return_if_fail (xml != NULL);
+
+	if (val_utf8 == NULL)
+		return;
+
 	if (id == NULL) {
-		xml->state = GSF_XML_OUT_CONTENT;
-		gsf_output_write (xml->output, 1, ">");
+		close_tag_if_neccessary (xml);
 	} else
 		gsf_output_printf (xml->output, " %s=\"", id);
 	while (*cur != '\0') {
@@ -889,6 +1148,15 @@ gsf_xml_out_add_cstr (GsfXMLOut *xml, char const *id,
 				gsf_output_write (xml->output, cur-start, start);
 			start = ++cur;
 			gsf_output_write (xml->output, 6, "&quot;");
+		} else if (*cur < 0x20 && id != NULL) {
+			guint8 buf[8];
+			sprintf (buf, "&#%d;", *cur);
+
+			if (cur != start)
+				gsf_output_write (xml->output, cur-start, start);
+			start = ++cur;
+
+			gsf_output_write (xml->output, strlen (buf), buf);
 		} else if (((*cur >= 0x20) && (*cur != 0x7f)) ||
 			   (*cur == '\n') || (*cur == '\r') || (*cur == '\t')) {
 			cur++;
@@ -950,7 +1218,7 @@ void
 gsf_xml_out_add_uint (GsfXMLOut *xml, char const *id,
 		      unsigned int val)
 {
-	char buf [4 * sizeof (int)];
+	char buf [4 * sizeof (unsigned int)];
 	sprintf (buf, "%u", val);
 	gsf_xml_out_add_cstr_unchecked (xml, id, buf);
 }
@@ -960,24 +1228,24 @@ gsf_xml_out_add_uint (GsfXMLOut *xml, char const *id,
  * @xml :
  * @id  : optionally NULL for content
  * @val : the value
- * @precision : the number of decimal points to display
+ * @precision : the number of significant digits to use, -1 meaning "enough".
  *
  * dump float value @val to an attribute named @id or as the nodes
- * content with precision @precision.
+ * content with precision @precision.  The number will be formattted
+ * according to the "C" locale.
  **/
 void
 gsf_xml_out_add_float (GsfXMLOut *xml, char const *id,
 		       double val, int precision)
 {
-	char buf [101 + DBL_DIG];
+	char format_str[4 * sizeof (int) + 10];
+	char buf[G_ASCII_DTOSTR_BUF_SIZE + DBL_DIG];
 
 	if (precision < 0 || precision > DBL_DIG)
 		precision = DBL_DIG;
 
-	if (fabs (val) < 1e9 && fabs (val) > 1e-5)
-		snprintf (buf, sizeof buf-1, "%.*g", precision, val);
-	else
-		snprintf (buf, sizeof buf-1, "%f", val);
+	sprintf (format_str, "%%.%dg", precision);
+	g_ascii_formatd (buf, sizeof (buf), format_str, val);
 	gsf_xml_out_add_cstr_unchecked (xml, id, buf);
 }
 
@@ -995,11 +1263,24 @@ void
 gsf_xml_out_add_color (GsfXMLOut *xml, char const *id,
 		       unsigned int r, unsigned int g, unsigned int b)
 {
-	char buf [4 * sizeof (unsigned)];
+	char buf [3 * 4 * sizeof (unsigned int) + 1];
 	sprintf (buf, "%X:%X:%X", r, g, b);
 	gsf_xml_out_add_cstr_unchecked (xml, id, buf);
 }
 
+void
+gsf_xml_out_add_enum (GsfXMLOut *xml, char const *id, GType etype, gint val)
+{
+	GEnumClass *eclass = G_ENUM_CLASS (g_type_class_peek (etype));
+	GEnumValue *ev = g_enum_get_value (eclass, val);
+
+	if (ev)
+		gsf_xml_out_add_cstr_unchecked (xml, id, ev->value_name);
+	else
+		g_warning ("Invalid value %d for type %s",
+			   val, g_type_name (etype));
+}
+
 /**
  * gsf_xml_out_add_base64 :
  * @xml :
@@ -1024,4 +1305,3 @@ gsf_xml_out_add_base64 (GsfXMLOut *xml, char const *id,
 	gsf_xml_out_add_cstr_unchecked (xml, id, tmp);
 	g_free (tmp);
 }
-
